//
//  Renderer.swift
//  PlanetHex
//
//  Created by Don McBrien on 23/03/2020.
//  Copyright © 2020 Don McBrien. All rights reserved.
//

import Foundation
import MetalKit

class Renderer: NSObject {
   static var device: MTLDevice!
   static var commandQueue: MTLCommandQueue!
   static var library: MTLLibrary!
   
   // Uniforms
   var vertexUniforms = VertexUniforms()
   //MARK:- temporary
   var outputBuffer: MTLBuffer!
   var outputBufferLength = 0
   //MARK:- end of temporary

   let renderPipelineState: MTLRenderPipelineState
   var renderEncoder: MTLRenderCommandEncoder!
   var depthStencilState: MTLDepthStencilState?
   let texture: MTLTexture!
   let samplerState: MTLSamplerState?
   
   var mesh: Mesh
   init(view: MTKView) {
      guard
         let device = MTLCreateSystemDefaultDevice(),
         let commandQueue = device.makeCommandQueue(),
         let library = device.makeDefaultLibrary()
         else {
            fatalError("GPU not available")
      }
      guard
         let url = Bundle.main.url(forResource: "Terrain",
                                   withExtension: "png")
         else {
            fatalError("Terrain.png not available")
      }

      Renderer.device = device
      Renderer.commandQueue = commandQueue
      Renderer.library = library
      depthStencilState = Renderer.buildDepthStencilState()!
      renderPipelineState = Renderer.buildPipelineState()
      mesh = Mesh()
      
      do {
         let textureLoaderOptions: [MTKTextureLoader.Option: Any] =
            [.origin: MTKTextureLoader.Origin.bottomLeft,
             .SRGB: false,
             .generateMipmaps: NSNumber(booleanLiteral: true)]
         let textureLoader = MTKTextureLoader(device: Renderer.device)
         texture = try textureLoader.newTexture(URL: url, options: textureLoaderOptions)
      } catch let error {
         fatalError(error.localizedDescription)
      }
      samplerState = Renderer.buildSamplerState()

      super.init()
      
      view.device = device
      view.depthStencilPixelFormat = .depth32Float
      view.clearColor = MTLClearColor(red: 0.3, green: 0.3, blue: 0.3, alpha: 1.0)
      view.delegate = self
      mtkView(view, drawableSizeWillChange: view.bounds.size)
   }
}

extension Renderer: MTKViewDelegate {
   func draw(in view: MTKView) {
      guard
         let renderPassDescriptor = view.currentRenderPassDescriptor,
         let commandBuffer = Renderer.commandQueue.makeCommandBuffer(),
         let encoder =
         commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor),
         let drawable = view.currentDrawable else {
            return
      }
      
      renderEncoder = encoder
      let clearColor = MTLClearColor(red: 0.3, green: 0.3, blue: 0.3, alpha: 1)
      renderPassDescriptor.colorAttachments[0].clearColor = clearColor
      renderEncoder.setCullMode(.back)
      renderEncoder.setDepthStencilState(depthStencilState)
      renderEncoder.setRenderPipelineState(renderPipelineState)
      renderEncoder.setTriangleFillMode(.fill)  // or .lines
      
      provideGPUWithData()
      
      renderEncoder.endEncoding()
      commandBuffer.present(drawable)
      commandBuffer.commit()
      
      //MARK:- temporary
      var array = [Output]()
      array.reserveCapacity(mesh.vertexCount)

      for i in 0..<mesh.vertexCount {
         array.append(outputBuffer.contents().load(fromByteOffset: MemoryLayout<Output>.stride * i, as: Output.self))
      }
      for output in array { print("\(output)") }
      print("returned outputBuffer of length: \(outputBufferLength)")

      //MARK:- end of temporary
   }
   
   func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
      print("resizing screen")
   }
}

//MARK:- temporary
struct Output: CustomStringConvertible {
   var position: float4 = float4(0,0,0,0)
   var colour: float4 = float4(0,0,0,1)
   var uvLocal: float2 = float2(0,0)
   var uvOffset: float2 = float2(0,0)
   var uv: float2 = float2(0,0)
   var vertexType: uint16 = 0
   var currentTriangle: uint16 = 0
   var index: uint16 = 0
   
   var description: String {
      let str1 = "position: \(position)\n"
      let str2 = "  colour: \(colour)\n"
      let str3 = " uvLocal: \(uvLocal)\n"
      let str4 = "uvOffset: \(uvOffset)\n"
      let str5 = "      uv: \(uv)\n"
      let str6 = "vertexType: \(vertexType)  currentTriangle: \(currentTriangle)  index: \(index)\n"
      return str1 + str2 + str3 + str4 + str5 + str6
   }
}
//MARK:- end of temporary

extension Renderer {
   private func provideGPUWithData() {
      //MARK:- temporary
      outputBufferLength = MemoryLayout<Output>.stride * mesh.vertexCount
      outputBuffer = Renderer.device.makeBuffer(bytes: [Output](repeating: Output(),
                                                                count: mesh.vertexCount),
                                                length: outputBufferLength,
                                                options: [])!
      print("outputBuffer set up, length: \(outputBufferLength)")
      renderEncoder.setVertexBuffer(outputBuffer, offset: 0, index: 4)
      //MARK:- end of temporary

      // Load vertices (Buffer[0])
      let vertexBuffer = Renderer.device.makeBuffer(bytes: mesh.vertices,
                                                    length: MemoryLayout<Vertex>.stride * mesh.vertexCount,
                                                    options: [])!
      renderEncoder.setVertexBuffer(vertexBuffer, offset: 0, index: 0)
      
      // Load localUVCoords (Buffer[2])
      var localUVBuffer = Renderer.device.makeBuffer(bytes: mesh.localUVCoords,
                                                     length: MemoryLayout<float2>.stride *  mesh.localUVCoords.count,
                                                     options: [])!
      renderEncoder.setVertexBytes(&localUVBuffer,
                                   length: MemoryLayout<float2>.stride * mesh.localUVCoords.count,
                                   index: 2)

      // Load indexTables (Buffer[3])
      var indexTablesBuffer = Renderer.device.makeBuffer(bytes: mesh.indexTables,
                                                     length: MemoryLayout<UInt16>.stride * 18,
                                                     options: [])!
      renderEncoder.setVertexBytes(&indexTablesBuffer,
                                   length: MemoryLayout<UInt16>.stride * 18,
                                   index: 3)

      for (i, submesh) in mesh.submeshes.enumerated() {
         // Load Uniforms 
         vertexUniforms.uvOffset = submesh.uvOffset
         vertexUniforms.currentTriangle = UInt16(i)
         vertexUniforms.colour = submesh.colour
         renderEncoder.setVertexBytes(&vertexUniforms,
                                      length: MemoryLayout<VertexUniforms>.stride,
                                      index: 1)
         
         // Load indices
         let indexBuffer = Renderer.device.makeBuffer(bytes: submesh.indices,
                                                       length: MemoryLayout<UInt16>.stride * submesh.indexCount,
                                                       options: [])!

         renderEncoder.setFragmentTexture(texture, index: 0)
         renderEncoder.setFragmentSamplerState(samplerState,
                                               index: 0)

         renderEncoder.drawIndexedPrimitives(type: .triangle,
                                             indexCount: submesh.indexCount,
                                             indexType: .uint16,
                                             indexBuffer: indexBuffer,
                                             indexBufferOffset: 0)
      }
   }
   
   private static func buildPipelineState() -> MTLRenderPipelineState {
      let library = Renderer.library
      
      let pipelineDescriptor = MTLRenderPipelineDescriptor()
      pipelineDescriptor.vertexFunction = library?.makeFunction(name: "vertex_main")
      pipelineDescriptor.fragmentFunction = library?.makeFunction(name: "fragment_main")
      pipelineDescriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
      pipelineDescriptor.depthAttachmentPixelFormat = .depth32Float
      
      var pipelineState: MTLRenderPipelineState
      do {
         pipelineState = try Renderer.device.makeRenderPipelineState(descriptor: pipelineDescriptor)
      } catch let error {
         fatalError(error.localizedDescription)
      }
      return pipelineState
   }
   
   private static func buildDepthStencilState() -> MTLDepthStencilState? {
      let descriptor = MTLDepthStencilDescriptor()
      descriptor.depthCompareFunction = .less
      descriptor.isDepthWriteEnabled = true
      return device.makeDepthStencilState(descriptor: descriptor)
   }

   private static func buildSamplerState() -> MTLSamplerState? {
     let descriptor = MTLSamplerDescriptor()
     descriptor.sAddressMode = .repeat
     descriptor.tAddressMode = .repeat
     let samplerState =
          Renderer.device.makeSamplerState(descriptor: descriptor)
     return samplerState
   }
}

